#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>

#define FILEOUT "liste.txt"
#define FILETMP "/tmp/tmpXXXXXX"

struct student {
	int matrnr;
	char name[24];
	char fach[16];
};

off_t getStudent(int fd, struct student *p, off_t begin, off_t end) {

	off_t current;

    current = ((end - begin) / (2 * sizeof(struct student))) *
        sizeof(struct student) + begin;

    lseek(fd, current, SEEK_SET);
    read(fd, p, sizeof(struct student));

	return current;
}

/* Iterative Implementierung einer binaeren Suche. */
off_t searchStudent(int fd, struct student s) {

	off_t begin, end, current;
	struct student t;

	begin = 0;
	end = lseek(fd, 0, SEEK_END);

	current = getStudent(fd, &t, begin, end);

	while ( (strcmp(s.name, t.name) != 0) && (begin != end) ) {
		if (strcmp(s.name, t.name) < 0) {
			end = current;
			current = getStudent(fd, &t, begin, end);
		}
		else {
			begin = current + sizeof(struct student);
			current = getStudent(fd, &t, begin, end);
		}
	}
	if (begin == end) {
		return begin;
	}
	else {
		return current;
	}
}

/* Rekursive Implementierung einer binaeren Suche. */
/* Achtung: begin und end jetzt als zusaetzliche Parameter... */
/*
off_t searchStudent(struct student s, int fd, off_t begin, off_t end) {
    off_t current;
    struct student t;

    if (begin == end) {
        return begin;
    }

	current = getStudent(fd, &t, begin, end);

    if (strcmp(s.name, t.name) == 0) {
        return current;
    }
    else if (strcmp(s.name, t.name) < 0) {
        return searchStudent(s, fd, begin, current);
    }
    else {
        return searchStudent(s, fd, current + sizeof(struct student), end);
    }
}
*/

void addStudent() {
	struct student s, t;
	int fd, fdtmp;
	off_t current, size;
	
	/* tmp muss veraenderbar sein!!! */
	char tmp[30] = FILETMP;

	/* Studentendaten */	
    printf("\n");
    printf("Matrikelnummer? ");
    scanf("%d", &(s.matrnr));
    printf("Name? ");
    scanf("%s", s.name);
    printf("Fach? ");
    scanf("%s", s.fach);
    printf("\n");

	/* Datei oeffnen und falls nicht existiert dann anlegen*/
    if ((fd = open(FILEOUT, O_RDWR | O_CREAT, 0644)) < 0) {
   	    perror("ERROR OPEN");
       	exit(1);
    }

	size = lseek(fd, 0, SEEK_END);
	
	/* Falls Datei leer Datensatz schreiben. */
	if (size == 0) {
		write(fd, &s, sizeof(struct student));
		return;
	}

	current = searchStudent(fd, s);

    lseek(fd, current, SEEK_SET);
    read(fd, &t, sizeof(struct student));

	/* Falls Datensatz bereits existiert... */
    if (strcmp(s.name, t.name) == 0) {
		printf("RECORD ALREADY EXISTS!\n");
        return;
    }
	/* andernfalls einfuegen... */
	else {

		/* Geschickter ist mkstemp, da hier die Datei */
		/* mit den Rechten 0600 angelegt und auch gleich */
		/* geoeffnet wird (fdtmp = mkstemp(tmp)). */

	    if (!mktemp(tmp)) {
	   	    perror("ERROR MKTEMP");
        	exit(1);
	    }
	    if ((fdtmp = open(tmp, O_RDWR | O_CREAT | O_TRUNC | O_EXCL, 0600)) < 0) { 
	       	perror("ERROR OPEN TMP");
	        exit(1);
		}

		unlink(tmp);

    	/* Aktuellen Datensatz in tmp schreiben. */
	    write(fdtmp, &s, sizeof(struct student));

	    lseek(fd, current, SEEK_SET);

    	/* Alle Datensaetze ab current lesen. */
    	while ( (read(fd, &t, sizeof(struct student)) != 0) ) {

        	/* und in tmp schreiben */
	        write(fdtmp, &t, sizeof(struct student));
    	}

	    /* An den Anfang von fdtmp */
    	lseek(fdtmp, 0, SEEK_SET);
	    /* Zur aktuellen Position von fd */
    	lseek(fd, current, SEEK_SET);

	     /* Alle Datensaetze von fdtmp nach fd schreiben */
    	while ( (read(fdtmp, &t, sizeof(struct student)) != 0) ) {
        	write(fd, &t, sizeof(struct student));
	    }
	    close(fdtmp);
	}
	close (fd);
}

void findStudent() {

	int fd;
	off_t current;
	struct student s, t;

    printf("\n");
    printf("Name? ");
    scanf("%s", s.name);
    printf("\n");

    if ((fd = open(FILEOUT, O_RDONLY)) < 0) {
        perror("ERROR OPEN");
		printf("ADD A STUDENT!\n");
		return;
    }

    current = searchStudent(fd, s);

    lseek(fd, current, SEEK_SET);
    read(fd, &t, sizeof(struct student));
    if (strcmp(s.name, t.name) != 0) {
		printf("RECORD NOT FOUND!\n");
	}
	else {
	    lseek(fd, current, SEEK_SET);
    	read(fd, &s, sizeof(struct student));
	    printf("%d, %s, %s\n", s.matrnr, s.name, s.fach);
	}

	close(fd);
}

void changeStudent() {

	int fd;
	off_t current;
	struct student s, t;

    printf("\n");
    printf("Name? ");
    scanf("%s", s.name);
    printf("\n");

    if ((fd = open(FILEOUT, O_RDWR)) < 0) {
        perror("ERROR OPEN");
		printf("ADD A STUDENT!\n");
		return;
    }

    current = searchStudent(fd, s);

    lseek(fd, current, SEEK_SET);
    read(fd, &t, sizeof(struct student));
    if (strcmp(s.name, t.name) != 0) {
		printf("RECORD NOT FOUND!\n");
	}
	else {
		/* Studentendaten */	
    	printf("\n");
	    printf("Matrikelnummer? ");
    	scanf("%d", &(s.matrnr));
	    printf("Name? ");
    	scanf("%s", s.name);
	    printf("Fach? ");
    	scanf("%s", s.fach);
	    printf("---\n");

	    lseek(fd, current, SEEK_SET);
    	write(fd, &s, sizeof(struct student));
	}

	close(fd);

}


/* deleteStudent kann auch einfacher mit der */
/* Funktion truncate implementiert werden... */
void deleteStudent() {
    struct student s, t;
    int fd, fdtmp;
    off_t current, size;

    /* tmp muss veraenderbar sein!!! */
    char tmp[30] = FILETMP;

    printf("\n");
    printf("Name? ");
    scanf("%s", s.name);
    printf("\n");

    if ((fd = open(FILEOUT, O_RDWR)) < 0) {
        perror("ERROR OPEN");
		printf("ADD A STUDENT!\n");
        return;
    }

    size = lseek(fd, 0, SEEK_END);

    /* Falls Datei leer. */
    if (size == 0) {
		printf("NO RECORDS EXIST!\n");
    }

    current = searchStudent(fd, s);

    lseek(fd, current, SEEK_SET);
    read(fd, &t, sizeof(struct student));

    /* Falls Datensatz nicht existiert... */
    if (strcmp(s.name, t.name) != 0) {
        printf("RECORD DOES NOT EXISTS!\n");
        return;
    }
    /* andernfalls loeschen... */
    else {

        if (!mktemp(tmp)) {
            perror("ERROR MKTEMP");
            exit(1);
        }
        if ((fdtmp = open(tmp, O_RDWR | O_CREAT | O_TRUNC | O_EXCL, 0600)) < 0) {
            perror("ERROR OPEN TMP");
            exit(1);
        }

        unlink(tmp);

        /* Alle Datensaetze lesen und bis auf den */
		/* gesuchten alle in tmp schreiben. */
        lseek(fd, 0, SEEK_SET);
        while ( (read(fd, &t, sizeof(struct student)) != 0) ) {
            if (strcmp(s.name, t.name) != 0) {
                write(fdtmp, &t, sizeof(struct student));
            }
        }
		close(fd);

		/* Datei neu anlegen... */
        if ((fd = open(FILEOUT, O_RDWR | O_TRUNC)) < 0) {
            perror("ERROR OPEN");
            exit(1);
        }

		/* Alle Datensaetze von tmp in fd schreiben. */
        lseek(fdtmp, 0, SEEK_SET);
        while ( (read(fdtmp, &t, sizeof(struct student)) != 0) ) {
            write(fd, &t, sizeof(struct student));
        }
	}
}

void printStudents() {
    struct student s;
    int fd;
    if ((fd = open(FILEOUT, O_RDONLY)) < 0) {
        perror("ERROR OPEN");
		printf("ADD A STUDENT!\n");
        return;
    }
    while ( (read(fd, &s, sizeof(struct student)) != 0) ) {
        printf("%d, %s, %s\n", s.matrnr, s.name, s.fach);
    }
    close(fd);
}

int main() {

    char choice;

    printf("(a)dd, (f)ind, (c)hange, (d)delete, (p)rint, (q)uit? ");
    while ((choice = getchar()) != 'q') {
        switch(choice) {
            case 'a': addStudent(); break;
            case 'f': findStudent(); break;
            case 'c': changeStudent(); break;
            case 'd': deleteStudent(); break;
            case 'p': printStudents(); break;
            case '\n': continue;
        }
        printf("(a)dd, (f)ind, (c)hange, (d)elete, (p)rint, (q)uit? ");
    }

	return 0;

}
